from pyodmongo import DbModel, Field, Id, MainBaseModel
from typing import ClassVar
from bson import ObjectId
from pyodmongo.engines.utils import consolidate_dict


def test_save_dict_is_correct():
    class Lv3(DbModel):
        attr_lv3_one: str = Field(alias="attrLv3One")
        attr_lv3_two: str
        _collection: ClassVar = "lv3"

    class Lv2(DbModel):
        attr_lv2_one: str = Field(alias="attrLv2One")
        attr_lv2_two: str
        lv3: Lv3 | Id = Field(alias="lv3Alias")
        _collection: ClassVar = "lv2"

    class Lv1(DbModel):
        attr_lv1_one: str = Field(alias="attrLv1One")
        attr_lv1_two: str
        lv2: Lv2 | Id
        lv2_list: list[Lv2]
        lv2_list_ref: list[Id | Lv2]
        lv3_ref: Id | Lv3 = Field(alias="lv3Ref")
        lv3_list_ref: list[Lv3 | Id]
        _collection: ClassVar = "lv1"

    class Lv1Son(Lv1):
        lv1_son_attr: str

    id_1 = "64e8fe13e6dcc2a63c365df4"
    id_2 = "64e8fe13e6dcc2a63c365df5"
    id_3 = "64e8fe13e6dcc2a63c365df6"
    id_4 = "64e8fe13e6dcc2a63c365df7"
    id_5 = "64e8fe13e6dcc2a63c365df8"

    obj_lv3_1 = Lv3(
        id=id_1, attr_lv3_one="valor_attr_lv3_one", attr_lv3_two="valor_attr_lv3_two"
    )

    obj_lv3_2 = Lv3(
        id=id_2, attr_lv3_one="valor_attr_lv3_one", attr_lv3_two="valor_attr_lv3_two"
    )

    obj_lv_2_1 = Lv2(
        id=id_3,
        attr_lv2_one="valor_attr_lv2_one",
        attr_lv2_two="valor_attr_lv2_two",
        lv3=obj_lv3_1,
    )

    obj_lv_2_2 = Lv2(
        id=id_4,
        attr_lv2_one="valor_attr_lv2_one",
        attr_lv2_two="valor_attr_lv2_two",
        lv3=obj_lv3_2,
    )

    obj_lv1_son = Lv1Son(
        id=id_5,
        lv1_son_attr="value_lv1_son_attr",
        attr_lv1_one="value_attr_lv1_one",
        attr_lv1_two="value_attr_lv1_two",
        lv2=obj_lv_2_1,
        lv2_list=[obj_lv_2_1, obj_lv_2_2],
        lv2_list_ref=[obj_lv_2_1, obj_lv_2_2],
        lv3_ref=obj_lv3_2,
        lv3_list_ref=[obj_lv3_2, obj_lv3_1],
    )

    obj_lv3_1_expected_dict = {
        "_id": ObjectId("64e8fe13e6dcc2a63c365df4"),
        "attrLv3One": "valor_attr_lv3_one",
        "attr_lv3_two": "valor_attr_lv3_two",
        "created_at": None,
        "updated_at": None,
    }

    obj_lv_2_1_expected_dict = {
        "_id": ObjectId("64e8fe13e6dcc2a63c365df6"),
        "attrLv2One": "valor_attr_lv2_one",
        "attr_lv2_two": "valor_attr_lv2_two",
        "created_at": None,
        "lv3Alias": ObjectId("64e8fe13e6dcc2a63c365df4"),
        "updated_at": None,
    }

    obj_lv1_son_expected_dict = {
        "_id": ObjectId("64e8fe13e6dcc2a63c365df8"),
        "attrLv1One": "value_attr_lv1_one",
        "attr_lv1_two": "value_attr_lv1_two",
        "created_at": None,
        "lv1_son_attr": "value_lv1_son_attr",
        "lv2": ObjectId("64e8fe13e6dcc2a63c365df6"),
        "lv2_list": [
            {
                "_id": ObjectId("64e8fe13e6dcc2a63c365df6"),
                "attrLv2One": "valor_attr_lv2_one",
                "attr_lv2_two": "valor_attr_lv2_two",
                "created_at": None,
                "lv3Alias": ObjectId("64e8fe13e6dcc2a63c365df4"),
                "updated_at": None,
            },
            {
                "_id": ObjectId("64e8fe13e6dcc2a63c365df7"),
                "attrLv2One": "valor_attr_lv2_one",
                "attr_lv2_two": "valor_attr_lv2_two",
                "created_at": None,
                "lv3Alias": ObjectId("64e8fe13e6dcc2a63c365df5"),
                "updated_at": None,
            },
        ],
        "lv2_list_ref": [
            ObjectId("64e8fe13e6dcc2a63c365df6"),
            ObjectId("64e8fe13e6dcc2a63c365df7"),
        ],
        "lv3Ref": ObjectId("64e8fe13e6dcc2a63c365df5"),
        "lv3_list_ref": [
            ObjectId("64e8fe13e6dcc2a63c365df5"),
            ObjectId("64e8fe13e6dcc2a63c365df4"),
        ],
        "updated_at": None,
    }

    assert consolidate_dict(obj=obj_lv3_1, dct={}) == obj_lv3_1_expected_dict
    assert consolidate_dict(obj=obj_lv_2_1, dct={}) == obj_lv_2_1_expected_dict
    assert consolidate_dict(obj=obj_lv1_son, dct={}) == obj_lv1_son_expected_dict


def test_save_dict_with_basemodel_reference():
    class BaseModelClass(MainBaseModel):
        attr_bm: str

    class DbModelClass(DbModel):
        attr_dbm: str
        bm: BaseModelClass
        _collection: ClassVar = "db_model_class"

    obj = DbModelClass(attr_dbm="attr_dbm", bm=BaseModelClass(attr_bm="attr_bm"))

    dct = consolidate_dict(obj=obj, dct={})
    assert dct == {
        "_id": None,
        "created_at": None,
        "updated_at": None,
        "attr_dbm": "attr_dbm",
        "bm": {"attr_bm": "attr_bm"},
    }


def test_field_with_union_more_than_two():
    class MyFirstClass(DbModel):
        attr_first: str = None
        _collection: ClassVar = "my_first_class"

    class MyClass(DbModel):
        email: str = None
        mfc: list[MyFirstClass | Id] | None = None
        _collection: ClassVar = "my_class"

    obj = MyClass(mfc=None)
    dct = consolidate_dict(obj=obj, dct={})
    assert dct == {
        "_id": None,
        "created_at": None,
        "updated_at": None,
        "email": None,
        "mfc": None,
    }


def test_multiples_nested_references():
    class MyType1(MainBaseModel):
        attr_type_1: str = None

    class MyType2(MainBaseModel):
        attr_type_2: str = None

    class MyType3(MainBaseModel):
        attr_type_3: str = None

    class DbMyType(DbModel):
        attr_db: str = None
        my_type: MyType1 | MyType2 | MyType3 | None = None
        _collection: ClassVar = "db_my_type"

    obj1 = DbMyType(my_type=MyType1())
    obj2 = DbMyType(my_type=MyType2())
    obj3 = DbMyType(my_type=MyType3())
    dct1 = {
        "_id": None,
        "attr_db": None,
        "created_at": None,
        "my_type": {"attr_type_1": None},
        "updated_at": None,
    }
    dct2 = {
        "_id": None,
        "attr_db": None,
        "created_at": None,
        "my_type": {"attr_type_2": None},
        "updated_at": None,
    }
    dct3 = {
        "_id": None,
        "attr_db": None,
        "created_at": None,
        "my_type": {"attr_type_3": None},
        "updated_at": None,
    }

    assert consolidate_dict(obj=obj1, dct={}) == dct1
    assert consolidate_dict(obj=obj2, dct={}) == dct2
    assert consolidate_dict(obj=obj3, dct={}) == dct3


def test_list_of_ids_with_none():
    class MyType1(DbModel):
        attr_type_1: str = None
        _collection: ClassVar = "my_type1"

    class DbMyType(DbModel):
        attr_db: str = None
        my_type: list[MyType1 | Id] = None
        _collection: ClassVar = "db_my_type"

    obj = DbMyType()
    dct_expected = {
        "_id": None,
        "created_at": None,
        "updated_at": None,
        "attr_db": None,
        "my_type": None,
    }
    assert consolidate_dict(obj=obj, dct={}) == dct_expected

    class DbMyType(DbModel):
        attr_db: str = None
        my_type: list[Id] = None
        _collection: ClassVar = "db_my_type"

    obj = DbMyType()
    assert consolidate_dict(obj=obj, dct={}) == dct_expected


def test_fields_with_reference():
    class Model1(DbModel):
        attr_1: str = None
        ref_id: Id
        list_ref_id: list[Id]
        _collection: ClassVar = "model1"

    class Model2(DbModel):
        attr_2: str = None
        ref_id: Model1 | Id
        list_ref_id: list[Model1 | Id]
        _collection: ClassVar = "model2"

    id_1 = str(ObjectId())
    id_2 = str(ObjectId())
    id_3 = str(ObjectId())
    obj_1 = Model1(ref_id=id_1, list_ref_id=[id_2, id_3])

    id_4 = str(ObjectId())
    id_5 = str(ObjectId())
    id_6 = str(ObjectId())
    obj_2 = Model2(ref_id=id_4, list_ref_id=[id_5, id_6])

    dct_expected_1 = {
        "_id": None,
        "created_at": None,
        "updated_at": None,
        "attr_1": None,
        "ref_id": ObjectId(id_1),
        "list_ref_id": [ObjectId(id_2), ObjectId(id_3)],
    }

    dct_expected_2 = {
        "_id": None,
        "created_at": None,
        "updated_at": None,
        "attr_2": None,
        "ref_id": ObjectId(id_4),
        "list_ref_id": [ObjectId(id_5), ObjectId(id_6)],
    }
    assert consolidate_dict(obj=obj_1, dct={}) == dct_expected_1
    assert consolidate_dict(obj=obj_2, dct={}) == dct_expected_2



from pyodmongo import DbModel, Id
from typing import ClassVar


class User(DbModel):
    username: str
    password: str
    _collection: ClassVar = "users"


class Product(DbModel):
    name: str
    price: float
    is_available: bool
    user: User | Id
    _collection: ClassVar = "products"



from pyodmongo import DbEngine, MainBaseModel, DbModel
from typing import ClassVar
from pyodmongo.queries import elem_match

engine = DbEngine(mongo_uri="mongodb://localhost:27017", db_name="my_db")


class Product(MainBaseModel):
    name: str
    price: float
    is_available: bool


class Order(DbModel):
    code: str
    products: list[Product]
    _collection: ClassVar = "orders"


query = elem_match(Product.name == "Box", Product.price == 50, field=Order.products)
box: Product = engine.find_one(Model=Product, query=query)



from fastapi import FastAPI, Request
from pyodmongo import DbModel, AsyncDbEngine
from pyodmongo.queries import mount_query_filter
from typing import ClassVar

app = FastAPI()
engine = AsyncDbEngine(mongo_uri="mongodb://localhost:27017", db_name="my_db")


class MyModel(DbModel):
    attr1: str
    attr2: str
    attr3: int
    _collection: ClassVar = "my_model"


@app.get("/", response_model=list[MyModel])
async def get_route(request: Request):
    query, sort = mount_query_filter(
        Model=MyModel,
        items=request.query_params._dict,
        initial_comparison_operators=[],
    )
    return await engine.find_many(Model=MyModel, query=query, sort=sort)



from typing import ClassVar
import pytest
import pytest_asyncio
from pyodmongo import (
    AsyncDbEngine,
    DbEngine,
    DbModel,
    MainBaseModel,
    DbResponse,
    ResponsePaginate,
    Id,
    Field,
)
from pydantic import BaseModel
from bson import ObjectId
import copy
from faker import Faker


fake = Faker()


@pytest_asyncio.fixture
async def drop_db(async_engine: AsyncDbEngine, engine: DbEngine):
    await async_engine._client.drop_database("pyodmongo_pytest")
    engine._client.drop_database("pyodmongo_pytest")
    yield
    await async_engine._client.drop_database("pyodmongo_pytest")
    engine._client.drop_database("pyodmongo_pytest")


@pytest.mark.asyncio
async def test_save_all(async_engine: AsyncDbEngine, engine: DbEngine):
    class MyClass0(DbModel):
        attr_0: str = Field(index=True)
        attr_1: int = Field(index=True)
        _collection: ClassVar = "my_class_0"

    class MyClass1(DbModel):
        attr_2: str = Field(index=True)
        attr_3: int = Field(index=True)
        _collection: ClassVar = "my_class_1"

    obj_0 = MyClass0(attr_0="zero", attr_1=0)
    obj_1 = MyClass0(attr_0="one", attr_1=1)
    obj_2 = MyClass1(attr_2="two", attr_3=2)
    obj_3 = MyClass1(attr_2="three", attr_3=3)

    response_0: dict[str, DbResponse] = await async_engine.save_all([obj_0, obj_2])
    response_1: dict[str, DbResponse] = engine.save_all([obj_1, obj_3])

    assert response_0["my_class_0"].upserted_count == 1
    assert response_0["my_class_0"].upserted_ids[0] == obj_0.id
    assert response_0["my_class_1"].upserted_count == 1
    assert response_0["my_class_1"].upserted_ids[0] == obj_2.id
    assert response_1["my_class_0"].upserted_count == 1
    assert response_1["my_class_0"].upserted_ids[0] == obj_1.id
    assert response_1["my_class_1"].upserted_count == 1
    assert response_1["my_class_1"].upserted_ids[0] == obj_3.id

    assert ObjectId.is_valid(obj_0.id)
    assert ObjectId.is_valid(obj_1.id)
    assert ObjectId.is_valid(obj_2.id)
    assert ObjectId.is_valid(obj_3.id)

    id_0 = copy.copy(obj_0.id)
    id_1 = copy.copy(obj_1.id)
    id_2 = copy.copy(obj_2.id)
    id_3 = copy.copy(obj_3.id)

    obj_0.attr_0 = "zero_zero"
    obj_1.attr_0 = "one_one"
    obj_2.attr_2 = "two_two"
    obj_3.attr_2 = "three_three"

    await async_engine.save_all([obj_1, obj_3])
    engine.save_all([obj_0, obj_2])

    assert obj_0.id == id_0
    assert obj_1.id == id_1
    assert obj_2.id == id_2
    assert obj_3.id == id_3


@pytest.mark.asyncio
async def test_save(async_engine: AsyncDbEngine, engine: DbEngine, drop_db):
    class MyClass0(DbModel):
        attr_0: str = Field(index=True)
        attr_1: int = Field(index=True)
        _collection: ClassVar = "my_class_0"

    class MyClass1(DbModel):
        attr_2: str = Field(index=True)
        attr_3: int = Field(index=True)
        _collection: ClassVar = "my_class_1"

    obj_0 = MyClass0(attr_0="zero", attr_1=0)
    obj_1 = MyClass1(attr_2="two", attr_3=2)

    response_0: DbResponse = await async_engine.save(obj_0)
    response_1: DbResponse = engine.save(obj_1)

    assert obj_0.id == response_0.upserted_ids[0]
    assert obj_1.id == response_1.upserted_ids[0]


@pytest.mark.asyncio
async def test_find(async_engine: AsyncDbEngine, engine: DbEngine, drop_db):
    class MyClass0(DbModel):
        attr_0: str = Field(index=True)
        attr_1: int = Field(index=True)
        _collection: ClassVar = "my_class_0"

    objs_0_49 = [MyClass0(attr_0=fake.name(), attr_1=n) for n in range(0, 50)]
    objs_50_99 = [MyClass0(attr_0=fake.name(), attr_1=n) for n in range(50, 100)]

    await async_engine.save_all(objs_0_49)
    engine.save_all(objs_50_99)

    obj_to_find_0_49: MyClass0 = copy.deepcopy(objs_0_49[24])
    obj_to_find_50_99: MyClass0 = copy.deepcopy(objs_50_99[24])

    obj_found_0_49 = await async_engine.find_one(
        Model=MyClass0, query=MyClass0.id == obj_to_find_0_49.id
    )
    obj_found_50_49 = engine.find_one(
        Model=MyClass0, query=MyClass0.id == obj_to_find_50_99.id
    )

    assert obj_found_0_49 == obj_to_find_0_49
    assert obj_found_50_49 == obj_to_find_50_99

    query_0 = (MyClass0.attr_1 > 10) & (MyClass0.attr_1 <= 20)
    query_1 = (MyClass0.attr_1 > 60) & (MyClass0.attr_1 <= 70)

    result_0: ResponsePaginate = await async_engine.find_many(
        Model=MyClass0, query=query_0, paginate=True, current_page=2, docs_per_page=2
    )
    result_1: ResponsePaginate = engine.find_many(
        Model=MyClass0, query=query_1, paginate=True, current_page=2, docs_per_page=2
    )
    assert result_0.current_page == 2
    assert result_0.page_quantity == 5
    assert result_0.docs_quantity == 10
    assert len(result_0.docs) == 2
    assert result_1.current_page == 2
    assert result_1.page_quantity == 5
    assert result_1.docs_quantity == 10
    assert len(result_1.docs) == 2


@pytest.mark.asyncio
async def test_delete(async_engine: AsyncDbEngine, engine: DbEngine, drop_db):
    class MyClass0(DbModel):
        attr_0: str = Field(index=True)
        attr_1: int = Field(index=True)
        _collection: ClassVar = "my_class_0"

    objs_0_49 = [MyClass0(attr_0=fake.name(), attr_1=n) for n in range(0, 50)]
    objs_50_99 = [MyClass0(attr_0=fake.name(), attr_1=n) for n in range(50, 100)]

    await async_engine.save_all(objs_0_49)
    engine.save_all(objs_50_99)

    query_0 = (MyClass0.attr_1 >= 0) & (MyClass0.attr_1 < 10)
    query_1 = (MyClass0.attr_1 >= 50) & (MyClass0.attr_1 < 60)

    response_0: DbResponse = await async_engine.delete(
        Model=MyClass0, query=query_0, delete_one=True
    )
    response_1: DbResponse = engine.delete(
        Model=MyClass0, query=query_1, delete_one=True
    )
    assert response_0.deleted_count == 1
    assert response_1.deleted_count == 1

    find_result_0 = await async_engine.find_many(Model=MyClass0)
    find_result_1 = engine.find_many(Model=MyClass0)
    assert len(find_result_0) == 98
    assert len(find_result_1) == 98

    response_0: DbResponse = await async_engine.delete(Model=MyClass0, query=query_0)
    response_1: DbResponse = engine.delete(Model=MyClass0, query=query_1)
    assert response_0.deleted_count == 9
    assert response_1.deleted_count == 9

    find_result_0 = await async_engine.find_many(Model=MyClass0)
    find_result_1 = engine.find_many(Model=MyClass0)

    assert len(find_result_0) == 80
    assert len(find_result_1) == 80


@pytest.mark.asyncio
async def test_db_field_population(
    async_engine: AsyncDbEngine, engine: DbEngine, drop_db
):
    class MyClassA(DbModel):
        attr_a_1: str
        attr_a_2: str
        _collection: ClassVar = "my_class_a"

    class MyClassB(DbModel):
        attr_b_1: str
        attr_obj_a: MyClassA | Id
        _collection: ClassVar = "my_class_b"

    class MyClassC(DbModel):
        attr_c_1: str
        attr_obj_b: MyClassB | Id
        _collection: ClassVar = "my_class_c"

    obj_a = MyClassA(attr_a_1="value_a_1", attr_a_2="value_a_2")
    engine.save(obj_a)
    obj_b = MyClassB(attr_b_1="value_b_1", attr_obj_a=obj_a)
    await async_engine.save(obj_b)
    obj_c = MyClassC(attr_c_1="value_c_1", attr_obj_b=obj_b)
    await async_engine.save(obj_c)

    populate_db_fields = [MyClassC.attr_obj_b]
    obj_found = await async_engine.find_one(
        Model=MyClassC, populate=True, populate_db_fields=populate_db_fields
    )
    obj_c.attr_obj_b.attr_obj_a = obj_a.id
    assert obj_found == obj_c


@pytest.mark.asyncio
async def test_error_save_base_model(
    async_engine: AsyncDbEngine, engine: DbEngine, drop_db
):
    class MyClassA(BaseModel):
        a1: str = "a1"

    class MyClassB(DbModel):
        b1: str = "b1"
        obj_a: MyClassA = MyClassA()
        _collection: ClassVar = "my_class_b"

    obj = MyClassB()

    with pytest.raises(
        TypeError,
        match="The MyClassA class inherits from Pydantic's BaseModel class. Try switching to PyODMongo's MainBaseModel class",
    ):
        await async_engine.save(obj)

    with pytest.raises(
        TypeError,
        match="The MyClassA class inherits from Pydantic's BaseModel class. Try switching to PyODMongo's MainBaseModel class",
    ):
        engine.save(obj)


@pytest.mark.asyncio
async def test_list_of_empty_objects(
    async_engine: AsyncDbEngine, engine: DbEngine, drop_db
):
    class A(DbModel):
        name: str = "A Name"
        code: str = "A Code"
        _collection: ClassVar = "a"

    class B(MainBaseModel):
        a: A | Id
        code: str = "B Code"
        cost: float = 100

    class C(DbModel):
        name: str = "name_1"
        b_list: list[B] | None = None
        _collection: ClassVar = "c"

    obj_c = C()
    engine.save(obj_c)
    obj_found = await async_engine.find_one(Model=C, populate=True)
    assert obj_found == obj_c


@pytest.mark.asyncio
async def test_field_population_with_main_base_model(
    async_engine: AsyncDbEngine, engine: DbEngine, drop_db
):
    class S(DbModel):
        s1: str = "s1"
        _collection: ClassVar = "s"

    class BC(DbModel):
        b1: str = "b1"
        s: S | Id
        _collection: ClassVar = "bc"

    class A(MainBaseModel):
        a1: str = "a1"
        bc: BC | Id

    class P(DbModel):
        p1: str = "p1"
        a: A
        _collection: ClassVar = "p"

    class O(DbModel):
        o1: str = "o1"
        p: P | Id
        _collection: ClassVar = "o"

    obj_s = S()
    await async_engine.save(obj_s)
    obj_bc = BC(s=obj_s)
    await async_engine.save(obj_bc)
    obj_a = A(bc=obj_bc)
    obj_p = P(a=obj_a)
    engine.save(obj_p)
    obj_o = O(p=obj_p)
    engine.save(obj_o)

    obj_found = engine.find_one(
        Model=O, populate=True, populate_db_fields=[O.p, A.bc, BC.s]
    )
    assert obj_found == obj_o



from pathlib import Path
import os


class Replace:
    _exemples_folder = Path().absolute() / "docs_utils_pre_build" / "examples"

    def __init__(self, folder_path: Path) -> None:
        self.folder_path: Path = folder_path

    def _all_exemples(self):
        return list(
            filter(lambda x: x.endswith(".py"), os.listdir(self._exemples_folder))
        )

    def _all_md_files_on_folder(self):
        return list(filter(lambda x: x.endswith(".md"), os.listdir(self.folder_path)))

    def resolve_replace(self):
        file_names = self._all_md_files_on_folder()
        examples = self._all_exemples()
        for file_name in file_names:
            file_path: Path = self.folder_path / file_name
            with open(file=file_path, mode="r+", encoding="utf-8") as md_file:
                current_md_content = md_file.read()
                for exemple_name in examples:
                    exemple_path: Path = self._exemples_folder / exemple_name
                    with open(file=exemple_path, mode="r", encoding="utf-8") as py_file:
                        content_py = py_file.read()
                        current_md_content = current_md_content.replace(
                            f"__{exemple_name}__", content_py
                        )
                md_file.seek(0)
                md_file.write(current_md_content)
                md_file.truncate()


en_folder_path = Path().absolute() / ".docs_tmp" / "en"
pt_folder_path = Path().absolute() / ".docs_tmp" / "pt-BR"

en_obj = Replace(folder_path=en_folder_path)
pt_obj = Replace(folder_path=pt_folder_path)

en_obj.resolve_replace()
pt_obj.resolve_replace()



from typing import Any
from bson import ObjectId
from pydantic import GetCoreSchemaHandler
from pydantic_core import core_schema, CoreSchema
from pydantic_core.core_schema import ValidationInfo, str_schema


class Id(str):
    """
    Extends the base string type to implement custom validation and serialization logic
    specifically for IDs that are expected to be in MongoDB's ObjectId format. This class
    ensures that IDs are valid ObjectIds and can be used consistently throughout the model
    where ID validation and serialization are necessary.

    Methods:
        validate(cls, v, _): Validates that the given value `v` is a valid ObjectId. If `v`
                             is None, it returns None. If `v` is invalid, it raises a
                             ValueError.
        serialization(cls, v): Converts a valid ObjectId `v` into a string for serialization.
                                If `v` is invalid, it raises a ValueError.
        __get_pydantic_core_schema__(cls, source_type, handler): Returns a schema for use
                                    with Pydantic, providing metadata about how to handle
                                    the serialization and validation of ObjectIds, ensuring
                                    that data adheres to the expected format both in Python
                                    and JSON.

    These methods are class methods, meaning they are called on the class rather than on
    instances of the class.
    """

    @classmethod
    def validate(cls, v, _: ValidationInfo):
        if v is None:
            return v
        if not ObjectId.is_valid(v):
            raise ValueError("invalid Id")
        return str(v)

    # @classmethod
    # def serialization(cls, v):
    #     if not ObjectId.is_valid(v):
    #         raise ValueError("invalid Id")
    #     return str(v)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:  # type: ignore
        return core_schema.json_or_python_schema(
            python_schema=core_schema.with_info_plain_validator_function(cls.validate),
            json_schema=str_schema(),
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda v: str(v) if ObjectId.is_valid(v) else v
            ),
        )



from pyodmongo import DbModel
from pydantic import field_validator, ConfigDict
from typing import ClassVar


def test_single_field_validator():
    class MyModel1(DbModel):
        attr1: str
        attr2: str
        model_config = ConfigDict()
        _collection: ClassVar = "Vrau"

        @field_validator("attr2")
        def validate_attr2(cls, value):
            return value + "_mod"

    class MyModel2(MyModel1):
        attr3: str
        attr4: str

    obj = MyModel1(attr1="value_one", attr2="value_two")
    obj2 = MyModel2(
        attr1="value_one_one",
        attr2="value_two_two",
        attr3="value_three",
        attr4="value_four",
    )

    assert obj.attr2 == "value_two_mod"
    assert obj2.attr2 == "value_two_two_mod"



from pyodmongo import DbModel, Field
from typing import ClassVar


class Product(DbModel):
    name: str = Field(index=True)
    code: str = Field(index=True, unique=True)
    description: str = Field(text_index=True, default_language="english")
    price: float
    product_type: str
    is_available: bool
    _collection: ClassVar = "products"



from pyodmongo import DbModel, Field, DbEngine, AsyncDbEngine
from pymongo import IndexModel, ASCENDING, DESCENDING
from typing import ClassVar
import pytest
import pytest_asyncio


class MyClass(DbModel):
    attr_0: str
    attr_1: str = Field(index=True)
    attr_2: str = Field(unique=True)
    attr_3: str = Field(text_index=True)
    attr_4: str = Field(index=True, text_index=True, default_language="portuguese")
    _collection: ClassVar = "my_class"


class MyClass2(DbModel):
    attr_5: str
    attr_6: str = Field(index=True)
    _collection: ClassVar = "my_class_2"


@pytest.fixture()
def sync_drop_collection(engine: DbEngine):
    engine._db[MyClass._collection].drop()
    yield
    engine._db[MyClass._collection].drop()


@pytest_asyncio.fixture()
async def async_drop_collection(async_engine: AsyncDbEngine):
    await async_engine._db[MyClass._collection].drop()
    yield
    await async_engine._db[MyClass._collection].drop()


def test_index_unique_text_index():
    assert MyClass.model_fields["attr_1"].json_schema_extra.get("index") or False
    assert not MyClass.model_fields["attr_1"].json_schema_extra.get("unique") or False
    assert (
        not MyClass.model_fields["attr_1"].json_schema_extra.get("text_index") or False
    )

    assert not MyClass.model_fields["attr_2"].json_schema_extra.get("index") or False
    assert MyClass.model_fields["attr_2"].json_schema_extra.get("unique") or False
    assert (
        not MyClass.model_fields["attr_2"].json_schema_extra.get("text_index") or False
    )

    assert not MyClass.model_fields["attr_3"].json_schema_extra.get("index") or False
    assert not MyClass.model_fields["attr_3"].json_schema_extra.get("unique") or False
    assert MyClass.model_fields["attr_3"].json_schema_extra.get("text_index") or False

    assert MyClass.model_fields["attr_4"].json_schema_extra.get("index") or False
    assert not MyClass.model_fields["attr_4"].json_schema_extra.get("unique") or False
    assert MyClass.model_fields["attr_4"].json_schema_extra.get("text_index") or False


def test_check_index_field():
    assert hasattr(MyClass, "_init_indexes")
    assert len(MyClass._init_indexes) == 3
    assert all(type(item) is IndexModel for item in MyClass._init_indexes)


def test_sync_check_index_in_db(sync_drop_collection, engine: DbEngine):
    obj = MyClass(
        attr_0="attr_0",
        attr_1="attr_1",
        attr_2="attr_2",
        attr_3="attr_3",
        attr_4="attr_4",
    )
    result = engine.save(obj)
    indexes_in_db = engine._db[MyClass._collection].index_information()
    assert "attr_0" not in indexes_in_db
    assert "attr_1" in indexes_in_db
    assert "attr_2" not in indexes_in_db
    assert "attr_3" not in indexes_in_db
    assert "attr_4" in indexes_in_db
    assert "attr_3" in indexes_in_db["texts"]["weights"]
    assert "attr_4" in indexes_in_db["texts"]["weights"]


@pytest.mark.asyncio
async def test_async_check_index_in_db(
    async_drop_collection, async_engine: AsyncDbEngine
):
    obj = MyClass(
        attr_0="attr_0",
        attr_1="attr_1",
        attr_2="attr_2",
        attr_3="attr_3",
        attr_4="attr_4",
    )
    result = await async_engine.save(obj)
    indexes_in_db = await async_engine._db[MyClass._collection].index_information()
    assert "attr_0" not in indexes_in_db
    assert "attr_1" in indexes_in_db
    assert "attr_2" not in indexes_in_db
    assert "attr_3" not in indexes_in_db
    assert "attr_4" in indexes_in_db
    assert "attr_3" in indexes_in_db["texts"]["weights"]
    assert "attr_4" in indexes_in_db["texts"]["weights"]
    assert indexes_in_db["texts"]["default_language"] == "portuguese"


def test_create_indexes_on_inheritance(sync_drop_collection, engine: DbEngine):
    class InheritanceClass(MyClass):
        attr_5: str = Field(index=True)

    obj = InheritanceClass(
        attr_0="attr_0",
        attr_1="attr_1",
        attr_2="attr_2",
        attr_3="attr_3",
        attr_4="attr_4",
        attr_5="attr_5",
    )

    result = engine.save(obj)
    indexes_in_db = engine._db[MyClass._collection].index_information()
    assert "attr_0" not in indexes_in_db
    assert "attr_1" in indexes_in_db
    assert "attr_2" not in indexes_in_db
    assert "attr_3" not in indexes_in_db
    assert "attr_4" in indexes_in_db
    assert "attr_5" in indexes_in_db
    assert "attr_3" in indexes_in_db["texts"]["weights"]
    assert "attr_4" in indexes_in_db["texts"]["weights"]


def test_create_indexes_on_two_inheritance(engine: DbEngine):
    class MyClass3(MyClass, MyClass2):
        attr_7: str = Field(index=True)

    engine._db[MyClass._collection].drop()
    obj = MyClass3(
        attr_0="attr_0",
        attr_1="attr_1",
        attr_2="attr_2",
        attr_3="attr_3",
        attr_4="attr_4",
        attr_5="attr_5",
        attr_6="attr_6",
        attr_7="attr_7",
    )
    result = engine.save(obj)
    indexes_in_db = engine._db[MyClass._collection].index_information()
    assert "attr_0" not in indexes_in_db
    assert "attr_1" in indexes_in_db
    assert "attr_2" not in indexes_in_db
    assert "attr_3" not in indexes_in_db
    assert "attr_4" in indexes_in_db
    assert "attr_5" not in indexes_in_db
    assert "attr_6" in indexes_in_db
    assert "attr_7" in indexes_in_db
    assert "attr_3" in indexes_in_db["texts"]["weights"]
    assert "attr_4" in indexes_in_db["texts"]["weights"]
    engine._db[MyClass._collection].drop()


def test_create_custom_indexes(engine: DbEngine):
    class MyClassCustomIndex(DbModel):
        attr_0: str = Field(default=None, index=True)
        attr_1: str = Field(default=None)
        _collection: ClassVar = "my_class_custom_index"
        _indexes: ClassVar = [
            IndexModel(
                [("attr_0", ASCENDING), ("attr_1", DESCENDING)],
                name="attr_0_and_attr_1",
            )
        ]

    engine._db[MyClassCustomIndex._collection].drop()

    obj = MyClassCustomIndex()
    engine.save(obj)
    indexes_in_db = engine._db[MyClassCustomIndex._collection].index_information()
    assert "attr_0_and_attr_1" in indexes_in_db
    assert "attr_0" not in indexes_in_db
    assert "attr_1" not in indexes_in_db
    assert len(indexes_in_db["attr_0_and_attr_1"]["key"]) == 2
    assert indexes_in_db["attr_0_and_attr_1"]["key"][0] == ("attr_0", 1)
    assert indexes_in_db["attr_0_and_attr_1"]["key"][1] == ("attr_1", -1)

    engine._db[MyClassCustomIndex._collection].drop()


@pytest.mark.asyncio
async def test_async_create_custom_indexes(async_engine: AsyncDbEngine):
    class MyClassCustomIndex(DbModel):
        attr_0: str = Field(default=None, index=True)
        attr_1: str = Field(default=None)
        _collection: ClassVar = "my_class_custom_index"
        _indexes: ClassVar = [
            IndexModel(
                [("attr_0", ASCENDING), ("attr_1", DESCENDING)],
                name="attr_0_and_attr_1",
            )
        ]

    await async_engine._db[MyClassCustomIndex._collection].drop()

    obj = MyClassCustomIndex()
    await async_engine.save(obj)
    indexes_in_db = await async_engine._db[
        MyClassCustomIndex._collection
    ].index_information()
    assert "attr_0_and_attr_1" in indexes_in_db
    assert "attr_0" not in indexes_in_db
    assert "attr_1" not in indexes_in_db
    assert len(indexes_in_db["attr_0_and_attr_1"]["key"]) == 2
    assert indexes_in_db["attr_0_and_attr_1"]["key"][0] == ("attr_0", 1)
    assert indexes_in_db["attr_0_and_attr_1"]["key"][1] == ("attr_1", -1)

    await async_engine._db[MyClassCustomIndex._collection].drop()



import pydantic
import platform
import sys
import pymongo
import motor
from pyodmongo import version


def test_version():
    info = {
        "PyODMongo version": "__VERSION__",
        "Pydantic version": pydantic.version.VERSION,
        "Pymongo version": pymongo.__version__,
        "Motor version": motor._version.version,
        "Python version": sys.version,
        "Platform": platform.platform(),
    }
    response_expected = "\n".join([f"{key}: {value}" for key, value in info.items()])
    assert response_expected == version.response



from pyodmongo import DbModel, AsyncDbEngine, Id
from typing import ClassVar
import asyncio

engine = AsyncDbEngine(mongo_uri="mongodb://localhost:27017", db_name="my_db")


class Customer(DbModel):
    name: str
    email: str
    _collection: ClassVar = "customers"


class Order(DbModel):
    customer: Customer | Id
    value: float
    _collection: ClassVar = "orders"


class OrdersByCustomers(DbModel):
    count: int
    total_value: float
    _collection: ClassVar = "orders"
    _pipeline: ClassVar = [
        {
            "$group": {
                "_id": "$customer",
                "count": {"$count": {}},
                "total_value": {"$sum": "$value"},
            }
        }
    ]


async def main():
    result: list[OrdersByCustomers] = await engine.find_many(Model=OrdersByCustomers)


asyncio.run(main())



from pydantic import BaseModel
from .db_field_info import DbField


class SortOperator(BaseModel):
    """
    Represents a sort operation in a PyODMongo query, allowing for the specification
    of sorting criteria on database fields. This model is used to construct the sorting
    part of a query, defining how the results should be ordered.

    Attributes:
        operators (tuple[tuple[DbField, int], ...]): A tuple of tuples, where each
                     inner tuple contains a DbField and an integer indicating the
                     sort order. The integer should be 1 for ascending order and -1
                     for descending order.

    Methods:
        to_dict(): Converts the sort operation to a dictionary format suitable for
                   use in a MongoDB query. Validates that the sort order values are
                   either 1 or -1.
    """

    operators: tuple[tuple[DbField, int], ...]

    def to_dict(self):
        dct = {}
        for operator in self.operators:
            if operator[1] not in (1, -1):
                raise ValueError("Only values 1 ascending and -1 descending are valid")
            dct[operator[0].path_str] = operator[1]
        return dct



from typing import Any
from ..services.model_init import (
    resolve_indexes,
    resolve_class_fields_db_info,
)
from pydantic import BaseModel
from pydantic._internal._model_construction import ModelMetaclass
from typing_extensions import dataclass_transform


@dataclass_transform(kw_only_default=True)
class PyOdmongoMeta(ModelMetaclass):
    """
    Metaclass for creating and configuring PyODMongo models. This metaclass extends
    the functionality of Pydantic's ModelMetaclass to include database-specific
    configurations and transformations.

    The primary responsibilities of this metaclass are:
    - Setting and tracking the `__pyodmongo_complete__` attribute to ensure proper
      initialization of database-related features.
    - Applying necessary transformations to class fields for database compatibility.
    - Providing custom attribute access behavior for dynamically generated attributes.

    Attributes:
        __pyodmongo_complete__ (bool): Indicator of whether the meta-level
                                       configuration is complete.

    Methods:
        __new__(cls, name, bases, namespace, **kwargs): Constructs a new class instance,
            applying database-specific initializations.
        __getattr__(cls, name): Custom attribute access method supporting dynamic
            attribute retrieval based on model fields.
    """

    def __new__(
        cls, name: str, bases: tuple[Any], namespace: dict, **kwargs: Any
    ) -> type:
        setattr(cls, "__pyodmongo_complete__", False)
        for base in bases:
            setattr(base, "__pyodmongo_complete__", False)

        cls: BaseModel = ModelMetaclass.__new__(cls, name, bases, namespace, **kwargs)

        setattr(cls, "__pyodmongo_complete__", True)
        for base in bases:
            setattr(base, "__pyodmongo_complete__", True)

        resolve_class_fields_db_info(cls=cls)
        return cls

    def __getattr__(cls, name: str):
        if cls.__dict__.get("__pyodmongo_complete__"):
            is_attr = name in cls.__dict__.get("model_fields").keys()
            if is_attr:
                return cls.__dict__.get(name + "__pyodmongo")
        ModelMetaclass.__getattr__(cls, name)


@dataclass_transform(kw_only_default=True)
class DbMeta(PyOdmongoMeta):
    """
    Metaclass for database model entities in a PyODMongo environment. It extends
    the functionality of the PyOdmongoMeta by applying specific behaviors and
    transformations related to database operations such as indexing, reference
    resolution, and initialization of database fields.

    The primary responsibilities of this metaclass are:
    - Constructing new class instances with additional database-specific adjustments
      and initializations.
    - Setting up pipelines for resolving references within the database context.
    - Configuring indexes for efficient database operations.

    Attributes:
        __pyodmongo_complete__ (bool): Attribute used to track the completion of
                                       the meta-level configuration.

    Methods:
        __new__(cls, name, bases, namespace, **kwargs): Constructs a new class instance,
            ensuring database-specific adjustments and initializations are applied.
    """

    def __new__(
        cls, name: str, bases: tuple[Any], namespace: dict, **kwargs: Any
    ) -> type:
        cls: BaseModel = PyOdmongoMeta.__new__(cls, name, bases, namespace, **kwargs)

        indexes = resolve_indexes(cls=cls)
        setattr(cls, "_init_indexes", indexes)
        return cls



import re
import requests
from datetime import datetime
from pathlib import Path
import shutil

release_notes_api_url = "https://api.github.com/repos/mauro-andre/pyodmongo/releases"
release_notes = requests.get(url=release_notes_api_url).json()

en_file_path = Path().absolute() / ".docs_tmp" / "en" / "release_notes.md"
pt_file_path = Path().absolute() / ".docs_tmp" / "pt-BR" / "release_notes.md"


def convert_text(input_text):
    pattern = r"(@[\w-]+) in (\S+)"

    def replace_link(match):
        name = match.group(1)
        url = match.group(2)
        return f'<a href="{url}" target="_blank">{name}</a>'

    converted_text = re.sub(pattern, replace_link, input_text)
    return converted_text


with open(en_file_path, "w") as file:
    file.write("# <center>Release notes</center> #\n\n")
    for release_note in release_notes:
        name = release_note.get("name")
        published_at = datetime.fromisoformat(
            release_note.get("published_at")
        ).strftime("%Y-%m-%d")
        body: str = release_note.get("body")
        body = convert_text(body)
        body = body.replace("\n", "\n\n")
        file.write("---\n")
        file.write(f"## {name} - {published_at}\n")
        file.write(f"{body}\n\n")

shutil.copy(en_file_path, pt_file_path)



from pyodmongo.services.verify_subclasses import is_subclass
from pyodmongo import DbModel, MainBaseModel
from pyodmongo.models.query_operators import (
    QueryOperator,
    LogicalOperator,
    _LogicalOperator,
    ComparisonOperator,
    ElemMatchOperator,
)


def test_is_subclass():
    assert is_subclass(class_to_verify=QueryOperator, subclass=QueryOperator) is True
    assert is_subclass(class_to_verify=LogicalOperator, subclass=QueryOperator) is True
    assert is_subclass(class_to_verify=_LogicalOperator, subclass=QueryOperator) is True
    assert (
        is_subclass(class_to_verify=ComparisonOperator, subclass=QueryOperator) is True
    )
    assert (
        is_subclass(class_to_verify=ElemMatchOperator, subclass=QueryOperator) is True
    )
    assert is_subclass(class_to_verify=DbModel, subclass=QueryOperator) is False
    assert is_subclass(class_to_verify=MainBaseModel, subclass=QueryOperator) is False



from pyodmongo import DbModel
from pydantic import BaseModel
from typing import ClassVar


class Address(BaseModel):
    street: str
    city: str
    state: str
    zip_code: str


class User(DbModel):
    username: str
    password: str
    address: Address
    _collection: ClassVar = "users"



from bson import ObjectId
from pydantic import BaseModel
from ..models.id_model import Id
from ..models.db_field_info import DbField
from ..services.reference_pipeline import resolve_reference_pipeline
from ..services.verify_subclasses import is_subclass


def consolidate_dict(obj: BaseModel, dct: dict):
    """
    Recursively consolidates the attributes of a Pydantic model into a dictionary,
    handling nested models, list fields, and references appropriately based on the model's
    field specifications.

    Args:
        obj (BaseModel): The Pydantic model instance from which to extract data.
        dct (dict): The dictionary into which data should be consolidated. This dictionary
                    is modified in-place.

    Returns:
        dict: The modified dictionary with model data consolidated into it, including handling
              of MongoDB ObjectId conversions and nested structures.

    Description:
        This function iterates over each field of the provided model instance, extracting the
        value and associated metadata. Depending on whether the field contains model data,
        is a list, or is a reference, the function appropriately processes the value to fit
        MongoDB storage requirements, including converting to ObjectIds where necessary.
        The function is recursive for nested models and lists of models.
    """
    for field, field_info in obj.model_fields.items():
        value = getattr(obj, field)
        try:
            db_field_info: DbField = getattr(obj.__class__, field)
        except AttributeError:
            if is_subclass(class_to_verify=obj.__class__, subclass=BaseModel):
                raise TypeError(
                    f"The {obj.__class__.__name__} class inherits from Pydantic's BaseModel class. Try switching to PyODMongo's MainBaseModel class"
                )
        alias = db_field_info.field_alias
        has_model_fields = db_field_info.has_model_fields
        is_list = db_field_info.is_list
        by_reference = db_field_info.by_reference
        if has_model_fields:
            if value is None:
                dct[alias] = None
                continue
            if not is_list:
                if by_reference:
                    try:
                        dct[alias] = ObjectId(value.id)
                    except AttributeError:
                        dct[alias] = ObjectId(value)
                else:
                    dct[alias] = {}
                    consolidate_dict(obj=value, dct=dct[alias])
            else:
                if by_reference:
                    try:
                        dct[alias] = [ObjectId(o.id) for o in value]
                    except AttributeError:
                        dct[alias] = [ObjectId(o) for o in value]
                else:
                    dct[alias] = []
                    for v in value:
                        obj_lst_elem = {}
                        consolidate_dict(obj=v, dct=obj_lst_elem)
                        dct[alias].append(obj_lst_elem)
        else:
            if db_field_info.field_type == Id and value is not None:
                if is_list:
                    dct[alias] = [ObjectId(o) for o in value if ObjectId.is_valid(o)]
                else:
                    dct[alias] = ObjectId(value) if ObjectId.is_valid(value) else value
            else:
                dct[alias] = value
    return dct


def mount_base_pipeline(
    Model,
    query: dict,
    sort: dict,
    populate: bool,
    populate_db_fields: list[DbField] | None,
):
    """
    Constructs a basic MongoDB aggregation pipeline based on the provided query, sort, and
    model settings, optionally including reference population stages.

    Args:
        Model (type[Model]): The model class defining the MongoDB collection and its aggregation logic.
        query (dict): MongoDB query dictionary to filter the documents.
        sort (dict): Dictionary defining the sorting order of the documents.
        populate (bool): Flag indicating whether to include reference population stages in the pipeline.

    Returns:
        list: The MongoDB aggregation pipeline configured with match, sort, and optionally population stages.

    Description:
        This function constructs a MongoDB aggregation pipeline using the provided model's
        internal pipeline stages and the specified query and sort parameters. If 'populate' is
        true, reference population stages defined in the model are included to resolve document
        references as part of the aggregation process.
    """
    match_stage = [{"$match": query}]
    sort_stage = [{"$sort": sort}] if sort != {} else []
    model_stage = Model._pipeline
    if model_stage != []:
        return match_stage + model_stage + sort_stage
    if populate:
        reference_stage = resolve_reference_pipeline(
            cls=Model, pipeline=[], populate_db_fields=populate_db_fields
        )
        return match_stage + reference_stage + sort_stage
    else:
        return match_stage + sort_stage



from __future__ import annotations as _annotations

import typing
from typing import Any

from pydantic_core import PydanticUndefined
from typing_extensions import Literal, Unpack

from pydantic.fields import AliasPath, AliasChoices, _EmptyKwargs
from pydantic import Field as Pydantic_Field


_Unset: Any = PydanticUndefined


def Field(  # noqa: C901
    default: Any = PydanticUndefined,
    *,
    default_factory: typing.Callable[[], Any] | None = _Unset,
    alias: str | None = _Unset,
    alias_priority: int | None = _Unset,
    validation_alias: str | AliasPath | AliasChoices | None = _Unset,
    serialization_alias: str | None = _Unset,
    title: str | None = _Unset,
    description: str | None = _Unset,
    examples: list[Any] | None = _Unset,
    exclude: bool | None = _Unset,
    discriminator: str | None = _Unset,
    json_schema_extra: dict[str, Any] | None = _Unset,
    frozen: bool | None = _Unset,
    validate_default: bool | None = _Unset,
    repr: bool = _Unset,
    init_var: bool | None = _Unset,
    kw_only: bool | None = _Unset,
    pattern: str | None = _Unset,
    strict: bool | None = _Unset,
    gt: float | None = _Unset,
    ge: float | None = _Unset,
    lt: float | None = _Unset,
    le: float | None = _Unset,
    multiple_of: float | None = _Unset,
    allow_inf_nan: bool | None = _Unset,
    max_digits: int | None = _Unset,
    decimal_places: int | None = _Unset,
    min_length: int | None = _Unset,
    max_length: int | None = _Unset,
    union_mode: Literal["smart", "left_to_right"] = _Unset,
    index: bool = False,
    unique: bool = False,
    text_index: bool = False,
    default_language: str = None,
    **extra: Unpack[_EmptyKwargs],
) -> Any:
    """Usage docs: https://docs.pydantic.dev/2.2/usage/fields

    Create a field for objects that can be configured.

    Used to provide extra information about a field, either for the model schema or complex validation. Some arguments
    apply only to number fields (`int`, `float`, `Decimal`) and some apply only to `str`.

    Note:
        - Any `_Unset` objects will be replaced by the corresponding value defined in the `_DefaultValues` dictionary. If a key for the `_Unset` object is not found in the `_DefaultValues` dictionary, it will default to `None`

    Args:
        default: Default value if the field is not set.
        default_factory: A callable to generate the default value, such as :func:`~datetime.utcnow`.
        alias: An alternative name for the attribute.
        alias_priority: Priority of the alias. This affects whether an alias generator is used.
        validation_alias: 'Whitelist' validation step. The field will be the single one allowed by the alias or set of
            aliases defined.
        serialization_alias: 'Blacklist' validation step. The vanilla field will be the single one of the alias' or set
            of aliases' fields and all the other fields will be ignored at serialization time.
        title: Human-readable title.
        description: Human-readable description.
        examples: Example values for this field.
        exclude: Whether to exclude the field from the model serialization.
        discriminator: Field name for discriminating the type in a tagged union.
        json_schema_extra: Any additional JSON schema data for the schema property.
        frozen: Whether the field is frozen.
        validate_default: Run validation that isn't only checking existence of defaults. This can be set to `True` or `False`. If not set, it defaults to `None`.
        repr: A boolean indicating whether to include the field in the `__repr__` output.
        init_var: Whether the field should be included in the constructor of the dataclass.
        kw_only: Whether the field should be a keyword-only argument in the constructor of the dataclass.
        strict: If `True`, strict validation is applied to the field.
            See [Strict Mode](../usage/strict_mode.md) for details.
        gt: Greater than. If set, value must be greater than this. Only applicable to numbers.
        ge: Greater than or equal. If set, value must be greater than or equal to this. Only applicable to numbers.
        lt: Less than. If set, value must be less than this. Only applicable to numbers.
        le: Less than or equal. If set, value must be less than or equal to this. Only applicable to numbers.
        multiple_of: Value must be a multiple of this. Only applicable to numbers.
        min_length: Minimum length for strings.
        max_length: Maximum length for strings.
        pattern: Pattern for strings.
        allow_inf_nan: Allow `inf`, `-inf`, `nan`. Only applicable to numbers.
        max_digits: Maximum number of allow digits for strings.
        decimal_places: Maximum number of decimal places allowed for numbers.
        union_mode: The strategy to apply when validating a union. Can be `smart` (the default), or `left_to_right`.
            See [Union Mode](../usage/types/unions.md#union-mode) for details.
        extra: Include extra fields used by the JSON schema.

            !!! warning Deprecated
                The `extra` kwargs is deprecated. Use `json_schema_extra` instead.

    Returns:
        A new [`FieldInfo`][pydantic.fields.FieldInfo], the return annotation is `Any` so `Field` can be used on
            type annotated fields without causing a typing error.
    """
    # Check deprecated and removed params from V1. This logic should eventually be removed.
    json_schema_extra = (
        {}
        if json_schema_extra == PydanticUndefined or json_schema_extra is None
        else json_schema_extra
    )
    json_schema_extra["index"] = index
    json_schema_extra["unique"] = unique
    json_schema_extra["text_index"] = text_index
    json_schema_extra["default_language"] = default_language
    return Pydantic_Field(
        default=default,
        default_factory=default_factory,
        alias=alias,
        alias_priority=alias_priority,
        validation_alias=validation_alias,
        serialization_alias=serialization_alias,
        title=title,
        description=description,
        examples=examples,
        exclude=exclude,
        discriminator=discriminator,
        json_schema_extra=json_schema_extra,
        frozen=frozen,
        validate_default=validate_default,
        repr=repr,
        init_var=init_var,
        kw_only=kw_only,
        pattern=pattern,
        strict=strict,
        gt=gt,
        ge=ge,
        lt=lt,
        le=le,
        multiple_of=multiple_of,
        allow_inf_nan=allow_inf_nan,
        max_digits=max_digits,
        decimal_places=decimal_places,
        min_length=min_length,
        max_length=max_length,
        union_mode=union_mode,
        **extra,
    )
